#!/bin/sh

EX_OK=0;
EX_IOERR=74


# Bring down two loopback interfaces
lo_down()
{
        ifconfig lo2 destroy
        ifconfig lo3 destroy
}

# Create two loopback interfaces and assign example addresses
lo_up()
{
	# Just in case they are still around
	lo_down

        ifconfig lo2 create
        ifconfig lo3 create
	
	if [ -n "$ipv6" ]
	then
		ifconfig lo2 inet6 "$1/128" mtu 1500 up
		ifconfig lo3 inet6 "$2/128" mtu 1500 up
	else
        	ifconfig lo2 "$1/32" mtu 1500 up
        	ifconfig lo3 "$2/32" mtu 1500 up
	fi
}

needroot()
{
	if [ $(id -u) -ne 0 ]
	then
		echo "not ok $no - $name # Only root may run this test"
		exit 1
	fi
}

# Run a test and display results
runtest()
{
	killall "$bin" > /dev/null 2>&1

	if ! make -f "Makefile.${bin}" > /dev/null 2>&1
	then
		echo "not ok $no - $name # make failed"
		exit 1
	fi

	# Finally run the test
	./$bin $testargs >/dev/null 2>&1

	if [ "$?" -eq "$expect" ]
	then
		echo "ok $no - $name"
	else
		echo "not ok $no - $name"
	fi
	lo_down > /dev/null 2>&1
}

# Setup loopback interfaces before starting the test program and bring them
# down afterwards.
test_with_loopback()
{
	# Make sure we run as root
	needroot 

	# Create loopback interfaces
	lo_up $client $server > /dev/null 2>&1
	
	# Run the atual test
	runtest

	# Destroy loopback interfaces
	lo_down > /dev/null 2>&1
}

no=$1
case "$no" in
	# The test passes if the UTO values are correctly exchanged.
	"1")
		name="Send UTO IPv4"
		bin="send_uto"
		addr="127.0.0.1"
		port="1296"
		expect=0		# Normal exit
		
		testargs="$addr $port"
		runtest
	;;
	# In this scenario the server imposes a very small timeout. After the
	# handshake the connection is broken and the client can't receive data
 	# (and thus can't acknowledge it). *
	# The server wait a period longer than UTO before trying to write again;
	# this triggers an EPIPE error which means the connection is broken.
	#
	# The test passes if the program exits abnormally with EX_IOERR.
	"2")
		name="Short UTO IPv4"
		bin="generic_uto"
		expect=$EX_IOERR

		# Test arguments
		client="192.0.2.2"
		server="192.0.2.3"
		port="1296"
		client_snd="0"
		client_rcv="1"
		server_snd="10"		# Server imposes 10 seconds timeout
		server_rcv="1"
		downtime="20"		# Downtime lasts 20 seconds

		testargs="$client $server $port $client_snd $client_rcv
		$server_snd $server_rcv $downtime"

		test_with_loopback
	;;

	# In this scenario the client requests a large UTO value and the server
	# accepts it. The network is unavailable for a period < UTO. After the
	# network picks up, the client should eventually receive the data and
	# acknowledge it.
	#
	# The test passes if the program exits normally.
	"3")
		name="Long UTO IPv4"
		bin="generic_uto"
		expect=0		# Normal exit

		# Test arguments
		client="192.0.2.2"
		server="192.0.2.3"
		port="1296"
		client_snd="1200"	# 20 minutes UTO
		client_rcv="1"
		server_snd="0"
		server_rcv="1"		# Server accepts UTO suggestion
		downtime="300"		# 5 minutes downtime

		testargs="$client $server $port $client_snd $client_rcv
		$server_snd $server_rcv $downtime"

		test_with_loopback
	;;
	# Same as test 1, with IPv6
	"4")
		name="Send UTO IPv6"
		bin="send_uto"
		addr="::1"
		port="1296"
		expect=0		# Normal exit

		testargs="$addr $port"
		runtest
	;;
	# Same as test 2, with IPv6
	"5")
		name="Short UTO IPv6"
		bin="generic_uto"
		ipv6="yes"
		expect=$EX_IOERR

		# Test arguments
		client="2001:db80:5::2"
		server="2001:db80:5::3"
		port="1296"
		client_snd="0"
		client_rcv="1"
		server_snd="10"		# Server imposes 10 seconds timeout
		server_rcv="1"
		downtime="20"		# Downtime lasts 20 seconds

		testargs="$client $server $port $client_snd $client_rcv
		$server_snd $server_rcv $downtime"

		test_with_loopback
	;;
	# Same as test 3, with IPv6
	"6")
		name="Long UTO IPv6"
		bin="generic_uto"
		ipv6="yes"
		expect=0		# Normal exit

		# Test arguments
		client="2001:db80:5::2"
		server="2001:db80:5::3"
		port="1296"
		client_snd="1200"	# Client requests 20 minutes UTO
		client_rcv="1"
		server_snd="0"
		server_rcv="1"		# Server accepts UTO suggestions
		downtime="300"		# 5 minutes downtime

		testargs="$client $server $port $client_snd $client_rcv
		$server_snd $server_rcv $downtime"

		test_with_loopback
	;;
	# In this scenario the server doesn't accept the suggestion of the
	# client so the connection should fail.
	#
	# The test passes if the program exits abnormally with EX_IOERR.
	"7")
		name="Refused UTO"
		bin="generic_uto"
		expect=$EX_IOERR

		# Test arguments
		client="192.0.2.2"
		server="192.0.2.3"
		port="1296"
		client_snd="3600"	# Client requests 1 hour
		client_rcv="1"
		server_snd="0"
		server_rcv="0"		# Server refuses UTO
		downtime="600"		# 10 minutes downtime

		testargs="$client $server $port $client_snd $client_rcv
		$server_snd $server_rcv $downtime"

		test_with_loopback
	;;
	# In this scenario the client requests one hour of UTO but the
	# server accepts only 300 seconds.
	#
	# The test passes if the program exits with EX_IOERR.
	"8")
		name="Partial UTO IPv6"
		bin="generic_uto"
		ipv6="yes"
		expect=$EX_IOERR

		# Test arguments
		client="2001:db80:5::2"
		server="2001:db80:5::3"
		port="1296"
		client_snd="3600"	# Client requests 1 hour
		client_rcv="1"
		server_snd="0"
		server_rcv="300"	# Server accepts max 5 minutes
		downtime="600"		# 10 minutes downtime

		testargs="$client $server $port $client_snd $client_rcv
		$server_snd $server_rcv $downtime"

		test_with_loopback
	;;
esac
