#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sysexits.h>

#include "test_utils.h"

/*
 * These are utility functions used by UTO tests.
 */


/*
 * Bind a socket to an address, using the SO_REUSEADDR socket option.
 */
int
tcp_bind(char *addr, char *port)
{
	int s, rc, yes = 1;
	struct addrinfo hints;
	struct addrinfo *result, *rp;

	memset(&hints, 0, sizeof(hints));
	hints.ai_canonname = NULL;
	hints.ai_addr = NULL;
	hints.ai_next = NULL;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_family = AF_UNSPEC;		/* Both v4 and v6.*/
	hints.ai_flags = AI_PASSIVE;

	/* Address lookup. */
	rc = getaddrinfo(addr, port, &hints, &result);
	if (rc)
		return -1;

	/* Loop until we can bind to something. */
	for (rp = result; rp != NULL; rp = rp->ai_next) {
		s = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
		if (s == -1)
			/* Try another result. */
			continue;

		if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &yes,
		    sizeof(yes))) {
			close(s);
			freeaddrinfo(result);
			return -1;
		}

		if (bind(s, rp->ai_addr, rp->ai_addrlen) == 0)
			/* Return bound socket. */
			break;
		
		/* Couldn't bind; try another result. */
		close(s);
	}
	freeaddrinfo(result);

	return (rp == NULL) ? -1 : s;
}

/*
 * Try to connect an existing socket to and address.
 */
int
tcp_connect(int s, char *addr, char *port)
{
	int rc;
	struct addrinfo hints;
	struct addrinfo *result, *rp;

	memset(&hints, 0, sizeof(hints));
	hints.ai_canonname = NULL;
	hints.ai_addr = NULL;
	hints.ai_next = NULL;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_family = AF_UNSPEC;		/* Both v4 and v6.*/

	/* Look up address. */
	rc = getaddrinfo(addr, port, &hints, &result);
	if (rc)
		return -1;

	/* Search for a suitable combination. */
	for (rp = result; rp != NULL; rp = rp->ai_next) {
		if (connect(s, rp->ai_addr, rp->ai_addrlen) == 0)
			/* Done connect, break. */
			break;

		/* Connect failed; trying next entry. */
	}
	freeaddrinfo(result);

	return (rp == NULL) ? -1 : s;
}

/*
 * Set UTO values on a socket.
 */
void
uto_set(int s, int send_uto, int recv_uto)
{
	if (setsockopt(s, IPPROTO_TCP, TCP_SNDUTO_TIMEOUT, &send_uto,
	    sizeof(send_uto)))
		err(EX_UNAVAILABLE, "setsockopt");

	if (setsockopt(s, IPPROTO_TCP, TCP_RCVUTO_TIMEOUT, &recv_uto,
	    sizeof(recv_uto)))
		err(EX_UNAVAILABLE, "setsockopt");
}

/*
 * Fetch UTO values from a socket.
 */
void
uto_get(int s, int *send_uto, int *recv_uto)
{
	socklen_t optlen;

	optlen = sizeof(*send_uto);
	if (getsockopt(s, IPPROTO_TCP, TCP_SNDUTO_TIMEOUT, send_uto, &optlen))
		err(EX_UNAVAILABLE, "setsockopt");

	optlen = sizeof(*recv_uto);
	if (getsockopt(s, IPPROTO_TCP, TCP_RCVUTO_TIMEOUT, recv_uto, &optlen))
		err(EX_UNAVAILABLE, "setsockopt");
}

/*
 * Create a socket, set UTO values, bind it locally, start listening.
 */
int
listen_uto(char *laddr, char *port, int send_uto, int recv_uto)
{
	int s;
	
	s = tcp_bind(laddr, port);
	if (s < 0)
		err(EX_UNAVAILABLE, "tcp_bind");
	uto_set(s, send_uto, recv_uto);

	if(listen(s, -1))
		err(EX_UNAVAILABLE, "listen");
	return s;
}

/*
 * Create a socket, set UTO values, bind it locally and attempt to connect.
 */
int
connect_uto(char *laddr, char *raddr, char *rport, int send_uto, int recv_uto)
{
	int s;
	
	s = tcp_bind(laddr, NULL);
	if (s < 0)
		err(EX_UNAVAILABLE, "tcp_bind");

	uto_set(s, send_uto, recv_uto);
	if (s != tcp_connect(s, raddr, rport))
		err(EX_UNAVAILABLE, "tcp_connect");

	return s;
}


/*
 * Simulate broken connectivity by diverting the traffic to an unlistened
 * port.
 *
 * XXX-CN This is a hack. The problem is that there's no easy way to filter
 * the traffic on loopback devices in a way that doesn't interfere with
 * syscalls.
 */
void
port_down(char *port)
{
	static char cmd[256];

	snprintf(cmd, sizeof(cmd), "ipfw add 00050 divert 5000 tcp from"
	" any %s to any", port);

	system(cmd);
}

/*
 * Remove the rule added by port_down, thus ending the downtime.
 */
void
port_up()
{
	system("ipfw delete 00050");
}


/*
 * Waits for connection and then brings down the specified network. Then it
 * tries to send data and waits for a specified time.
 *
 * After waiting is over it brings the route back up and tries to send again.
 * If the connection is dead, write(2) should trigger an EPIPE.
 */
void
server_disconnect(char *laddr, char *lport, int downtime, int send_uto,
    int recv_uto)
{
	int listen_sock;
	int sock;
	static char buf[6000];
	ssize_t bytes;

	/* Bind and start listening. */
	listen_sock = listen_uto(laddr, lport, send_uto, recv_uto);

	sock = accept(listen_sock, NULL, NULL);
	if (sock < 0)
		err(EX_UNAVAILABLE, "accept");

	/* Take the link down and then attempt to send data. */
	if (downtime)
		port_down(lport);

	bytes = send(sock, &buf, sizeof(buf), MSG_NOSIGNAL);
	if (bytes < 0)
		err(EX_IOERR, "send");

	/* Sleep through downtime and then bring the link back up. */
	if (downtime) {
		sleep(downtime);
		port_up();
	}

	/* Try again. If it succeeds, the connection is still up. */
	bytes = send(sock, &buf, sizeof(buf), MSG_NOSIGNAL);
	if (bytes < 0)
		err(EX_IOERR, "send");
}

/*
 * Generic client that simply connects and tries to receive data.
 */
void
client_generic(char *laddr, char *raddr, char *rport, int send_uto, int recv_uto)
{
	int sock;
	static char buf[6000];
	ssize_t bytes;

	/* Connect to the server, with the specified UTO values. */
	sock = connect_uto(laddr, raddr, rport, send_uto, recv_uto);

	bytes = recv(sock, &buf, sizeof(buf), MSG_WAITALL);
	if (bytes < 0 || (unsigned) bytes < sizeof(buf))
		err(EX_IOERR, "send");
}
